# pylint: disable=too-many-arguments
""" example project for storj-python binding shows how to use binding for various tasks. """

from datetime import datetime

from .errors import StorjException, BucketNotEmptyError, BucketNotFoundError
from .module_classes import ListObjectsOptions, Permission, SharePrefix
from .uplink import Uplink

if __name__ == "__main__":

    # Storj configuration information
    MY_API_KEY = "change-me-to-the-api-key-created-in-satellite-gui"
    MY_SATELLITE = "us-central-1.tardigrade.io:7777"
    MY_BUCKET = "my-first-bucket"
    MY_STORJ_UPLOAD_PATH = "(optional): path / (required): filename"
    # (path + filename) OR filename
    MY_ENCRYPTION_PASSPHRASE = "you'll never guess this"

    # Source and destination path and file name for testing
    SRC_FULL_FILENAME = "filename with extension of source file on local system"
    DESTINATION_FULL_FILENAME = "filename with extension to save on local system"

    # try-except block to catch any storj exception
    try:
        # create an object of Uplink class
        uplink = Uplink()

        # function calls
        # request access using passphrase
        print("\nRequesting Access using passphrase...")
        access = uplink.request_access_with_passphrase(MY_SATELLITE, MY_API_KEY,
                                                       MY_ENCRYPTION_PASSPHRASE)
        print("Request Access: SUCCESS!")
        #

        # open Storj project
        print("\nOpening the Storj project, corresponding to the parsed Access...")
        project = access.open_project()
        print("Desired Storj project: OPENED!")
        #

        # enlist all the buckets in given Storj project
        print("\nListing bucket's names and creation time...")
        bucket_list = project.list_buckets()
        for bucket in bucket_list:
            # as python class object
            print(bucket.name, " | ", datetime.fromtimestamp(bucket.created))
            # as python dictionary
            print(bucket.get_dict())
        print("Buckets listing: COMPLETE!")
        #

        # delete given bucket
        print("\nDeleting '" + MY_BUCKET + "' bucket...")
        try:
            bucket = project.delete_bucket(MY_BUCKET)
        # if delete bucket fails due to "not empty", delete all the objects and try again
        except BucketNotEmptyError as exception:
            print("Error while deleting bucket: ", exception.message)
            print("Deleting object's inside bucket and try to delete bucket again...")
            # list objects in given bucket recursively using ListObjectsOptions
            print("Listing and deleting object's inside bucket...")
            objects_list = project.list_objects(MY_BUCKET, ListObjectsOptions(recursive=True))
            # iterate through all objects path
            for obj in objects_list:
                # delete selected object
                print("Deleting '" + obj.key)
                _ = project.delete_object(MY_BUCKET, obj.key)
            print("Delete all objects inside the bucket : COMPLETE!")

            # try to delete given bucket
            print("Deleting '" + MY_BUCKET + "' bucket...")
            _ = project.delete_bucket(MY_BUCKET)
            print("Desired bucket: DELETED")
        except BucketNotFoundError as exception:
            print("Desired bucket delete error: ", exception.message)
        #

        # create bucket in given project
        print("\nCreating '" + MY_BUCKET + "' bucket...")
        _ = project.create_bucket(MY_BUCKET)
        print("Desired Bucket: CREATED!")

        # as an example of 'put' , lets read and upload a local file
        # upload file/object
        print("\nUploading data...")
        # get handle of file to be uploaded
        file_handle = open(SRC_FULL_FILENAME, 'r+b')
        # get upload handle to specified bucket and upload file path
        upload = project.upload_object(MY_BUCKET, MY_STORJ_UPLOAD_PATH)
        #
        # upload file on storj
        upload.write_file(file_handle)
        #
        # commit the upload
        upload.commit()
        # close file handle
        file_handle.close()
        print("Upload: COMPLETE!")
        #

        # list objects in given bucket with above options or None
        print("\nListing object's names...")
        objects_list = project.list_objects(MY_BUCKET, ListObjectsOptions(recursive=True,
                                                                          system=True))
        # print all objects path
        for obj in objects_list:
            print(obj.key, " | ", obj.is_prefix)  # as python class object
            print(obj.get_dict())  # as python dictionary
        print("Objects listing: COMPLETE!")
        #

        # as an example of 'get' , lets download an object and write it to a local file
        # download file/object
        print("\nDownloading data...")
        # get handle of file which data has to be downloaded
        file_handle = open(DESTINATION_FULL_FILENAME, 'w+b')
        # get download handle to specified bucket and object path to be downloaded
        download = project.download_object(MY_BUCKET, MY_STORJ_UPLOAD_PATH)
        #
        # download data from storj to file
        download.read_file(file_handle)
        #
        # close the download stream
        download.close()
        # close file handle
        file_handle.close()
        print("Download: COMPLETE!")
        #

        # as an example of how to create shareable Access for easy storj access without
        # API key and Encryption PassPhrase
        # create new Access with permissions
        print("\nCreating new Access...")
        # set permissions for the new access to be created
        permissions = Permission(allow_list=True, allow_delete=False)
        # set shared prefix as list of dictionaries for the new access to be created
        shared_prefix = [SharePrefix(bucket=MY_BUCKET, prefix="")]
        # create new access
        new_access = access.share(permissions, shared_prefix)
        print("New Access: CREATED!")
        #

        # generate serialized access to share
        print("\nGenerating serialized Access...")
        serialized_access = access.serialize()
        print("Serialized shareable Access: ", serialized_access)
        #

        #
        # close given project using handle
        print("\nClosing Storj project...")
        project.close()
        print("Project CLOSED!")
        #

        #
        # as an example of how to retrieve information from shareable Access for storj access
        # retrieving Access from serialized Access
        print("\nParsing serialized Access...")
        shared_access = uplink.parse_access(serialized_access)
        print("Parsing Access: COMPLETE")
        #

        # open Storj project
        print("\nOpening the Storj project, corresponding to the shared Access...")
        shared_project = shared_access.open_project()
        print("Desired Storj project: OPENED!")
        #

        # enlist all the buckets in given Storj project
        print("\nListing bucket's names and creation time...")
        bucket_list = shared_project.list_buckets()
        for bucket in bucket_list:
            # as python class object
            print(bucket.name, " | ", datetime.fromtimestamp(bucket.created))
            # as python dictionary
            print(bucket.get_dict())
        print("Buckets listing: COMPLETE!")
        #

        # list objects in given bucket with above options or None
        print("\nListing object's names...")
        objects_list = shared_project.list_objects(MY_BUCKET, ListObjectsOptions(recursive=True,
                                                                                 system=True))
        # print all objects path
        for obj in objects_list:
            print(obj.key, " | ", obj.is_prefix)  # as python class object
            print(obj.get_dict())  # as python dictionary
        print("Objects listing: COMPLETE!")
        #

        # try to delete given bucket
        print("\nTrying to delete '" + MY_STORJ_UPLOAD_PATH)
        try:
            _ = shared_project.delete_object(MY_BUCKET, MY_STORJ_UPLOAD_PATH)
            print("Desired object: DELETED")
        except StorjException as exception:
            print("Desired object: FAILED")
            print("Exception: ", exception.details)
        #

        #
        # close given project with shared Access
        print("\nClosing Storj project...")
        shared_project.close()
        print("Project CLOSED!")
        #
    except StorjException as exception:
        print("Exception Caught: ", exception.details)



"""Module with Upload class and upload methods to work with object upload"""
# pylint: disable=line-too-long
import ctypes
import os

from uplink_python.module_classes import CustomMetadata
from uplink_python.module_def import _UploadStruct, _WriteResult, _Error, _CustomMetadataStruct, _ObjectResult
from uplink_python.errors import _storj_exception

_WINDOWS = os.name == 'nt'
COPY_BUFSIZE = 1024 * 1024 if _WINDOWS else 64 * 1024


class Upload:
    """
    Upload is an upload to Storj Network.

    ...

    Attributes
    ----------
    upload : int
        Upload _handle returned from libuplinkc upload_result.upload
    uplink : Uplink
        uplink object used to get access

    Methods
    -------
    write():
        Int
    write_file():
        None
    commit():
        None
    abort():
        None
    set_custom_metadata():
        None
    info():
        Object
    """

    def __init__(self, upload, uplink):
        """Constructs all the necessary attributes for the Upload object."""

        self.upload = upload
        self.uplink = uplink

    def write(self, data_to_write: bytes, size_to_write: int):
        """
        function uploads bytes data passed as parameter to the object's data stream.

        Parameters
        ----------
        data_to_write : bytes
        size_to_write : int

        Returns
        -------
        int
        """

        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.upload_write.argtypes = [ctypes.POINTER(_UploadStruct),
                                                         ctypes.POINTER(ctypes.c_uint8),
                                                         ctypes.c_size_t]
        self.uplink.m_libuplink.upload_write.restype = _WriteResult
        #
        # prepare the inputs for the function
        # --------------------------------------------
        # data conversion to type required by function
        # get size of data in c type int32 variable
        # conversion of read bytes data to c type ubyte Array
        data_to_write = (ctypes.c_uint8 * ctypes.c_int32(len(data_to_write)).value)(*data_to_write)
        # conversion of c type ubyte Array to LP_c_ubyte required by upload write function
        data_to_write_ptr = ctypes.cast(data_to_write, ctypes.POINTER(ctypes.c_uint8))
        # --------------------------------------------
        size_to_write_obj = ctypes.c_size_t(size_to_write)

        # upload data by calling the exported golang function
        write_result = self.uplink.m_libuplink.upload_write(self.upload, data_to_write_ptr,
                                                            size_to_write_obj)
        #
        # if error occurred
        if bool(write_result.error):
            _storj_exception(write_result.error.contents.code,
                             write_result.error.contents.message.decode("utf-8"))
        return int(write_result.bytes_written)

    def write_file(self, file_handle, buffer_size: int = 0):
        """
        function uploads complete file whose handle is passed as parameter to the
        object's data stream and commits the object after upload is complete.

        Note: File handle should be a BinaryIO, i.e. file should be opened using 'r+b" flag.
        e.g.: file_handle = open(SRC_FULL_FILENAME, 'r+b')
        Remember to commit the object on storj and also close the local file handle
        after this function exits.

        Parameters
        ----------
        file_handle : BinaryIO
        buffer_size : int

        Returns
        -------
        None
        """

        if not buffer_size:
            buffer_size = COPY_BUFSIZE
        while True:
            buf = file_handle.read(buffer_size)
            if not buf:
                break
            self.write(buf, len(buf))

    def commit(self):
        """
        function commits the uploaded data.

        Returns
        -------
        None
        """

        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.upload_commit.argtypes = [ctypes.POINTER(_UploadStruct)]
        self.uplink.m_libuplink.upload_commit.restype = ctypes.POINTER(_Error)
        #

        # upload commit by calling the exported golang function
        error = self.uplink.m_libuplink.upload_commit(self.upload)
        #
        # if error occurred
        if bool(error):
            raise _storj_exception(error.contents.code,
                                   error.contents.message.decode("utf-8"))

    def abort(self):
        """
        function aborts an ongoing upload.

        Returns
        -------
        None
        """
        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.upload_abort.argtypes = [ctypes.POINTER(_UploadStruct)]
        self.uplink.m_libuplink.upload_abort.restype = ctypes.POINTER(_Error)
        #

        # abort ongoing upload by calling the exported golang function
        error = self.uplink.m_libuplink.upload_abort(self.upload)
        #
        # if error occurred
        if bool(error):
            raise _storj_exception(error.contents.code,
                                   error.contents.message.decode("utf-8"))

    def set_custom_metadata(self, custom_metadata: CustomMetadata = None):
        """
        function to set custom meta information while uploading data

        Parameters
        ----------
        custom_metadata : CustomMetadata

        Returns
        -------
        None
        """
        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.upload_set_custom_metadata.argtypes = [ctypes.POINTER(_UploadStruct),
                                                                       _CustomMetadataStruct]
        self.uplink.m_libuplink.upload_set_custom_metadata.restype = ctypes.POINTER(_Error)
        #
        # prepare the input for the function
        if custom_metadata is None:
            custom_metadata_obj = _CustomMetadataStruct()
        else:
            custom_metadata_obj = custom_metadata.get_structure()
        #
        # set custom metadata to upload by calling the exported golang function
        error = self.uplink.m_libuplink.upload_set_custom_metadata(self.upload, custom_metadata_obj)
        #
        # if error occurred
        if bool(error):
            raise _storj_exception(error.contents.code,
                                   error.contents.message.decode("utf-8"))

    def info(self):
        """
        function returns the last information about the uploaded object.

        Returns
        -------
        Object
        """
        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.upload_info.argtypes = [ctypes.POINTER(_UploadStruct)]
        self.uplink.m_libuplink.upload_info.restype = _ObjectResult
        #
        # get last upload info by calling the exported golang function
        object_result = self.uplink.m_libuplink.upload_info(self.upload)
        #
        # if error occurred
        if bool(object_result.error):
            raise _storj_exception(object_result.error.contents.code,
                                   object_result.error.contents.message.decode("utf-8"))
        return self.uplink.object_from_result(object_result.object)



# pylint: disable=too-few-public-methods, too-many-arguments
"""Python Binding's Uplink Module for Storj (V3)"""

import ctypes
import os
import sysconfig

from uplink_python.access import Access
from uplink_python.errors import _storj_exception, LibUplinkSoError
from uplink_python.module_def import _AccessResult, _ConfigStruct
from uplink_python.module_classes import Config, Bucket, Object, SystemMetadata,\
    CustomMetadataEntry, CustomMetadata


class Uplink:
    """
    Python Storj Uplink class to initialize and get access grant to Storj (V3)"

    ...

    Attributes
    ----------
    m_libuplink : CDLL
        Instance to the libuplinkc.so.

    Methods
    -------
    object_from_result(object_=object_result.object):
        Object
    bucket_from_result(bucket=bucket_result.bucket):
        Bucket
    """

    __instance = None

    def __init__(self):
        """Constructs all the necessary attributes for the Uplink object."""
        # private members of PyStorj class with reference objects
        # include the golang exported libuplink library functions
        if Uplink.__instance is None:
            so_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'libuplinkc.so')
            if os.path.exists(so_path):
                self.m_libuplink = ctypes.CDLL(so_path)
            else:
                new_path = os.path.join(sysconfig.get_paths()['purelib'], "uplink_python",
                                        'libuplinkc.so')
                if os.path.exists(new_path):
                    self.m_libuplink = ctypes.CDLL(so_path)
                else:
                    raise LibUplinkSoError
            Uplink.__instance = self
        else:
            self.m_libuplink = Uplink.__instance.m_libuplink

    @classmethod
    def object_from_result(cls, object_):
        """Converts ctypes structure _ObjectStruct to python class object."""

        system = SystemMetadata(created=object_.contents.system.created,
                                expires=object_.contents.system.expires,
                                content_length=object_.contents.system.content_length)

        array_size = object_.contents.custom.count
        entries = list()
        for i in range(array_size):
            if bool(object_.contents.custom.entries[i]):
                entries_obj = object_.contents.custom.entries[i]
                entries.append(CustomMetadataEntry(key=entries_obj.key.decode("utf-8"),
                                                   key_length=entries_obj.key_length,
                                                   value=entries_obj.value.decode("utf-8"),
                                                   value_length=entries_obj.value_length))
            else:
                entries.append(CustomMetadataEntry())

        return Object(key=object_.contents.key.decode("utf-8"),
                      is_prefix=object_.contents.is_prefix,
                      system=system,
                      custom=CustomMetadata(entries=entries,
                                            count=object_.contents.custom.count))

    @classmethod
    def bucket_from_result(cls, bucket_):
        """Converts ctypes structure _BucketStruct to python class object."""

        return Bucket(name=bucket_.contents.name.decode("utf-8"),
                      created=bucket_.contents.created)

    #
    def request_access_with_passphrase(self, satellite: str, api_key: str, passphrase: str):
        """
        RequestAccessWithPassphrase generates a new access grant using a passhprase.
        It must talk to the Satellite provided to get a project-based salt for deterministic
        key derivation.

        Note: this is a CPU-heavy function that uses a password-based key derivation
        function (Argon2). This should be a setup-only step.
        Most common interactions with the library should be using a serialized access grant
        through ParseAccess directly.

        Parameters
        ----------
        satellite : str
        api_key : str
        passphrase : str

        Returns
        -------
        Access
        """
        #
        # declare types of arguments and response of the corresponding golang function
        self.m_libuplink.request_access_with_passphrase.argtypes = [ctypes.c_char_p,
                                                                    ctypes.c_char_p,
                                                                    ctypes.c_char_p]
        self.m_libuplink.request_access_with_passphrase.restype = _AccessResult
        #
        # prepare the input for the function
        satellite_ptr = ctypes.c_char_p(satellite.encode('utf-8'))
        api_key_ptr = ctypes.c_char_p(api_key.encode('utf-8'))
        passphrase_ptr = ctypes.c_char_p(passphrase.encode('utf-8'))

        # get access to Storj by calling the exported golang function
        access_result = self.m_libuplink.request_access_with_passphrase(satellite_ptr,
                                                                        api_key_ptr,
                                                                        passphrase_ptr)
        #
        # if error occurred
        if bool(access_result.error):
            raise _storj_exception(access_result.error.contents.code,
                                   access_result.error.contents.message.decode("utf-8"))
        return Access(access_result.access, self)

    def config_request_access_with_passphrase(self, config: Config, satellite: str, api_key: str,
                                              passphrase: str):
        """
        RequestAccessWithPassphrase generates a new access grant using a passhprase and
        custom configuration.
        It must talk to the Satellite provided to get a project-based salt for deterministic
        key derivation.

        Note: this is a CPU-heavy function that uses a password-based key derivation
        function (Argon2). This should be a setup-only step.
        Most common interactions with the library should be using a serialized access grant
        through ParseAccess directly.

        Parameters
        ----------
        config: Config
        satellite : str
        api_key : str
        passphrase : str

        Returns
        -------
        Access
        """

        #
        # declare types of arguments and response of the corresponding golang function
        self.m_libuplink.config_request_access_with_passphrase.argtypes = [_ConfigStruct,
                                                                           ctypes.c_char_p,
                                                                           ctypes.c_char_p,
                                                                           ctypes.c_char_p]
        self.m_libuplink.config_request_access_with_passphrase.restype = _AccessResult
        #
        # prepare the input for the function
        if config is None:
            config_obj = _ConfigStruct()
        else:
            config_obj = config.get_structure()
        satellite_ptr = ctypes.c_char_p(satellite.encode('utf-8'))
        api_key_ptr = ctypes.c_char_p(api_key.encode('utf-8'))
        passphrase_ptr = ctypes.c_char_p(passphrase.encode('utf-8'))

        # get access to Storj by calling the exported golang function
        access_result = self.m_libuplink.config_request_access_with_passphrase(config_obj,
                                                                               satellite_ptr,
                                                                               api_key_ptr,
                                                                               passphrase_ptr)
        #
        # if error occurred
        if bool(access_result.error):
            raise _storj_exception(access_result.error.contents.code,
                                   access_result.error.contents.message.decode("utf-8"))
        return Access(access_result.access, self)

    def parse_access(self, serialized_access: str):
        """
        ParseAccess parses a serialized access grant string.

        This should be the main way to instantiate an access grant for opening a project.
        See the note on RequestAccessWithPassphrase

        Parameters
        ----------
        serialized_access : str

        Returns
        -------
        Access
        """

        #
        # prepare the input for the function
        serialized_access_ptr = ctypes.c_char_p(serialized_access.encode('utf-8'))
        #
        # declare types of arguments and response of the corresponding golang function
        self.m_libuplink.parse_access.argtypes = [ctypes.c_char_p]
        self.m_libuplink.parse_access.restype = _AccessResult
        #

        # get parsed access by calling the exported golang function
        access_result = self.m_libuplink.parse_access(serialized_access_ptr)
        #
        # if error occurred
        if bool(access_result.error):
            raise _storj_exception(access_result.error.contents.code,
                                   access_result.error.contents.message.decode("utf-8"))
        return Access(access_result.access, self)



"""Module with Download class and dowload methods to work with object download"""
# pylint: disable=too-many-arguments
import ctypes
import os

from uplink_python.module_def import _DownloadStruct, _ReadResult, _ProjectStruct, _ObjectResult, _Error
from uplink_python.errors import _storj_exception

_WINDOWS = os.name == 'nt'
COPY_BUFSIZE = 1024 * 1024 if _WINDOWS else 64 * 1024


class Download:
    """
    Download is a download from Storj Network.

    ...

    Attributes
    ----------
    download : int
        Download _handle returned from libuplinkc download_result.download
    uplink : Uplink
        uplink object used to get access
    project : Project
        project object used to create upload
    bucket_name : Str
        bucket_name to which upload is being processed
    storj_path : Str
        storj_path on which upload is to be done

    Methods
    -------
    read():
        Int
    read_file():
        None
    file_size():
        Int
    close():
        None
    info():
        Object
    """

    def __init__(self, download, uplink, project, bucket_name, storj_path):
        """Constructs all the necessary attributes for the Download object."""

        self.download = download
        self.project = project
        self.bucket_name = bucket_name
        self.storj_path = storj_path
        self.uplink = uplink

    def read(self, size_to_read: int):
        """
        function downloads up to len size_to_read bytes from the object's data stream.
        It returns the data_read in bytes and number of bytes read

        Parameters
        ----------
        size_to_read : int

        Returns
        -------
        bytes, int
        """
        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.download_read.argtypes = [ctypes.POINTER(_DownloadStruct),
                                                          ctypes.POINTER(ctypes.c_uint8),
                                                          ctypes.c_size_t]
        self.uplink.m_libuplink.download_read.restype = _ReadResult
        #
        # prepare the inputs for the function
        data_size = ctypes.c_int32(size_to_read)
        data_to_write = [0]
        data_to_write = (ctypes.c_uint8 * data_size.value)(*data_to_write)
        data_to_write_ptr = ctypes.cast(data_to_write, ctypes.POINTER(ctypes.c_uint8))
        size_to_read = ctypes.c_size_t(size_to_read)

        # read data from Storj by calling the exported golang function
        read_result = self.uplink.m_libuplink.download_read(self.download, data_to_write_ptr,
                                                            size_to_read)
        #
        # if error occurred
        if bool(read_result.error):
            raise _storj_exception(read_result.error.contents.code,
                                   read_result.error.contents.message.decode("utf-8"))

        data_read = bytes()
        if int(read_result.bytes_read) != 0:
            #
            # --------------------------------------------
            # data conversion to type python readable form
            # conversion of LP_c_ubyte to python readable data variable
            data_read = ctypes.string_at(data_to_write_ptr, int(read_result.bytes_read))
        return data_read, int(read_result.bytes_read)

    def read_file(self, file_handle, buffer_size: int = 0):
        """
        function downloads complete object from it's data stream and writes it to the file whose
        handle is passed as parameter. After the download is complete it closes the download stream.

        Note: File handle should be a BinaryIO, i.e. file should be opened using 'w+b" flag.
        e.g.: file_handle = open(DESTINATION_FULL_FILENAME, 'w+b')
        Remember to close the object stream on storj and also close the local file handle
        after this function exits.

        Parameters
        ----------
        file_handle : BinaryIO
        buffer_size : int

        Returns
        -------
        None
        """
        if not buffer_size:
            buffer_size = COPY_BUFSIZE
        file_size = self.file_size()
        if buffer_size > file_size:
            buffer_size = file_size
        while file_size:
            buf, bytes_read = self.read(buffer_size)
            if not buf:
                file_handle.write(buf)
            file_size -= bytes_read

    def file_size(self):
        """
        function returns the size of object on Storj network for which download has been created.

        Returns
        -------
        int
        """

        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.stat_object.argtypes = [ctypes.POINTER(_ProjectStruct),
                                                        ctypes.c_char_p, ctypes.c_char_p]
        self.uplink.m_libuplink.stat_object.restype = _ObjectResult
        #
        # get object information by calling the exported golang function
        object_result = self.uplink.m_libuplink.stat_object(self.project, self.bucket_name,
                                                            self.storj_path)
        # if error occurred
        if bool(object_result.error):
            raise _storj_exception(object_result.error.contents.code,
                                   object_result.error.contents.message.decode("utf-8"))
        # find object size
        return int(object_result.object.contents.system.content_length)

    def close(self):
        """
        function closes the download.

        Returns
        -------
        None
        """
        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.close_download.argtypes = [ctypes.POINTER(_DownloadStruct)]
        self.uplink.m_libuplink.close_download.restype = ctypes.POINTER(_Error)
        #
        # close downloader by calling the exported golang function
        error = self.uplink.m_libuplink.close_download(self.download)
        #
        # if error occurred
        if bool(error):
            raise _storj_exception(error.contents.code,
                                   error.contents.message.decode("utf-8"))

    def info(self):
        """
        function returns information about the downloaded object.

        Returns
        -------
        Object
        """
        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.download_info.argtypes = [ctypes.POINTER(_DownloadStruct)]
        self.uplink.m_libuplink.download_info.restype = _ObjectResult
        #
        # get last download info by calling the exported golang function
        object_result = self.uplink.m_libuplink.download_info(self.download)
        #
        # if error occurred
        if bool(object_result.error):
            raise _storj_exception(object_result.error.contents.code,
                                   object_result.error.contents.message.decode("utf-8"))
        return self.uplink.object_from_result(object_result.object)



"""Module with Access class and access methods to get access grant to access project"""
import ctypes

from uplink_python.module_classes import Permission, SharePrefix, Config
from uplink_python.module_def import _ConfigStruct, _PermissionStruct, _SharePrefixStruct, \
    _AccessStruct, _ProjectResult, _StringResult, _AccessResult
from uplink_python.project import Project
from uplink_python.errors import _storj_exception


class Access:
    """
    An Access Grant contains everything to access a project and specific buckets.
    It includes a potentially-restricted API Key, a potentially-restricted set of encryption
    information, and information about the Satellite responsible for the project's metadata.

    ...

    Attributes
    ----------
    access : int
        Access _handle returned from libuplinkc access_result.access
    uplink : Uplink
        uplink object used to get access

    Methods
    -------
    open_project():
        Project
    config_open_project():
        Project
    serialize():
        String
    share():
        Access
    """

    def __init__(self, access, uplink):
        """Constructs all the necessary attributes for the Access object."""

        self.access = access
        self.uplink = uplink

    def open_project(self):
        """
        function opens Storj(V3) project using access grant.

        Returns
        -------
        Project
        """

        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.open_project.argtypes = [ctypes.POINTER(_AccessStruct)]
        self.uplink.m_libuplink.open_project.restype = _ProjectResult
        #
        # open project by calling the exported golang function
        project_result = self.uplink.m_libuplink.open_project(self.access)
        #
        # if error occurred
        if bool(project_result.error):
            raise _storj_exception(project_result.error.contents.code,
                                   project_result.error.contents.message.decode("utf-8"))
        return Project(project_result.project, self.uplink)

    def config_open_project(self, config: Config):
        """
        function opens Storj(V3) project using access grant and custom configuration.

        Parameters
        ----------
        config : Config

        Returns
        -------
        Project
        """

        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.config_open_project.argtypes = [_ConfigStruct,
                                                                ctypes.POINTER(_AccessStruct)]
        self.uplink.m_libuplink.config_open_project.restype = _ProjectResult
        #
        # prepare the input for the function
        if config is None:
            config_obj = _ConfigStruct()
        else:
            config_obj = config.get_structure()
        #
        # open project by calling the exported golang function
        project_result = self.uplink.m_libuplink.config_open_project(config_obj, self.access)
        #
        # if error occurred
        if bool(project_result.error):
            raise _storj_exception(project_result.error.contents.code,
                                   project_result.error.contents.message.decode("utf-8"))
        return Project(project_result.project, self.uplink)

    def serialize(self):
        """
        function serializes an access grant such that it can be used later
        with ParseAccess or other tools.

        Returns
        -------
        String
        """

        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.access_serialize.argtypes = [ctypes.POINTER(_AccessStruct)]
        self.uplink.m_libuplink.access_serialize.restype = _StringResult
        #
        # get serialized access by calling the exported golang function
        string_result = self.uplink.m_libuplink.access_serialize(self.access)
        #
        # if error occurred
        if bool(string_result.error):
            raise _storj_exception(string_result.error.contents.code,
                                   string_result.error.contents.message.decode("utf-8"))
        return string_result.string.decode("utf-8")

    def share(self, permission: Permission = None, shared_prefix: [SharePrefix] = None):
        """
        function Share creates a new access grant with specific permissions.

        Access grants can only have their existing permissions restricted, and the resulting
        access grant will only allow for the intersection of all previous Share calls in the
        access grant construction chain.

        Prefixes, if provided, restrict the access grant (and internal encryption information)
        to only contain enough information to allow access to just those prefixes.

        Parameters
        ----------
        permission : Permission
        shared_prefix : list of SharePrefix

        Returns
        -------
        Access
        """

        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.access_share.argtypes = [ctypes.POINTER(_AccessStruct),
                                                         _PermissionStruct,
                                                         ctypes.POINTER(_SharePrefixStruct),
                                                         ctypes.c_size_t]
        self.uplink.m_libuplink.access_share.restype = _AccessResult
        #
        # prepare the input for the function
        # check and create valid _PermissionStruct parameter
        if permission is None:
            permission_obj = _PermissionStruct()
        else:
            permission_obj = permission.get_structure()

        # check and create valid Share Prefix parameter
        if shared_prefix is None:
            shared_prefix_obj = ctypes.POINTER(_SharePrefixStruct)()
            array_size = ctypes.c_size_t(0)
        else:
            num_of_structs = len(shared_prefix)
            li_array_size = (_SharePrefixStruct * num_of_structs)()
            array = ctypes.cast(li_array_size, ctypes.POINTER(_SharePrefixStruct))
            for i, val in enumerate(shared_prefix):
                array[i] = val.get_structure()
            shared_prefix_obj = array
            array_size = ctypes.c_size_t(num_of_structs)
        #
        # get shareable access by calling the exported golang function
        access_result = self.uplink.m_libuplink.access_share(self.access, permission_obj,
                                                             shared_prefix_obj, array_size)
        #
        # if error occurred
        if bool(access_result.error):
            raise _storj_exception(access_result.error.contents.code,
                                   access_result.error.contents.message.decode("utf-8"))
        return Access(access_result.access, self.uplink)



# pylint: disable=missing-docstring
import unittest

from .test_data.access_test import AccessTest
from .test_data.bucket_list_test import BucketListTest
from .test_data.bucket_test import BucketTest
from .test_data.helper import InitializationTest
from .test_data.object_list_test import ObjectListTest
from .test_data.object_test import ObjectTest
from .test_data.project_test import ProjectTest

if __name__ == '__main__':
    testList = [InitializationTest, AccessTest, ProjectTest, BucketTest, BucketListTest,
                ObjectTest, ObjectListTest]
    testLoad = unittest.TestLoader()

    TestList = []
    for testCase in testList:
        testSuite = testLoad.loadTestsFromTestCase(testCase)
        TestList.append(testSuite)

    newSuite = unittest.TestSuite(TestList)
    runner = unittest.TextTestRunner(verbosity=4)
    runner.run(newSuite)



"""Classes for input and output interface of parameters and returns from uplink."""
# pylint: disable=too-few-public-methods, too-many-arguments
import ctypes

from uplink_python.module_def import _ConfigStruct, _PermissionStruct, _SharePrefixStruct, _BucketStruct,\
    _DownloadOptionsStruct, _SystemMetadataStruct, _CustomMetadataStruct, _UploadOptionsStruct,\
    _ObjectStruct, _ListObjectsOptionsStruct, _ListBucketsOptionsStruct, _CustomMetadataEntryStruct


class Config:
    """
    Config defines configuration for using uplink library.

    ...

    Attributes
    ----------
    user_agent : str
    dial_timeout_milliseconds : int
        DialTimeout defines how long client should wait for establishing a connection to peers.
    temp_directory : str
        temp_directory specifies where to save data during downloads to use less memory.

    Methods
    -------
    get_structure():
        _ConfigStruct
    """

    def __init__(self, user_agent: str = "", dial_timeout_milliseconds: int = 0,
                 temp_directory: str = ""):
        """Constructs all the necessary attributes for the Config object."""

        self.user_agent = user_agent
        self.dial_timeout_milliseconds = dial_timeout_milliseconds
        self.temp_directory = temp_directory

    def get_structure(self):
        """Converts python class object to ctypes structure _ConfigStruct"""

        return _ConfigStruct(ctypes.c_char_p(self.user_agent.encode('utf-8')),
                             ctypes.c_int32(self.dial_timeout_milliseconds),
                             ctypes.c_char_p(self.temp_directory.encode('utf-8')))


class Permission:
    """
    Permission defines what actions can be used to share.

    ...

    Attributes
    ----------
    allow_download : bool
        allow_download gives permission to download the object's content. It
        allows getting object metadata, but it does not allow listing buckets.
    allow_upload : bool
        allow_upload gives permission to create buckets and upload new objects.
        It does not allow overwriting existing objects unless allow_delete is
        granted too.
    allow_list : bool
        allow_list gives permission to list buckets. It allows getting object
        metadata, but it does not allow downloading the object's content.
    allow_delete : bool
        allow_delete gives permission to delete buckets and objects. Unless
        either allow_download or allow_list is granted too, no object metadata and
        no error info will be returned for deleted objects.
    not_before : int
        NotBefore restricts when the resulting access grant is valid for.
        If set, the resulting access grant will not work if the Satellite
        believes the time is before NotBefore.
        If set, this value should always be before NotAfter.
        disabled when 0.
    not_after : int
        NotAfter restricts when the resulting access grant is valid till.
        If set, the resulting access grant will not work if the Satellite
        believes the time is after NotAfter.
        If set, this value should always be after NotBefore.
        disabled when 0.

    Methods
    -------
    get_structure():
        _PermissionStruct
    """

    def __init__(self, allow_download: bool = False, allow_upload: bool = False,
                 allow_list: bool = False, allow_delete: bool = False,
                 not_before: int = 0, not_after: int = 0):
        """Constructs all the necessary attributes for the Permission object."""

        self.allow_download = allow_download
        self.allow_upload = allow_upload
        self.allow_list = allow_list
        self.allow_delete = allow_delete
        self.not_before = not_before
        self.not_after = not_after

    def get_structure(self):
        """Converts python class object to ctypes structure _PermissionStruct"""

        return _PermissionStruct(ctypes.c_bool(self.allow_download),
                                 ctypes.c_bool(self.allow_upload),
                                 ctypes.c_bool(self.allow_list),
                                 ctypes.c_bool(self.allow_delete),
                                 ctypes.c_int64(self.not_before),
                                 ctypes.c_int64(self.not_after))


class SharePrefix:
    """
    SharePrefix defines a prefix that will be shared.

    ...

    Attributes
    ----------
    bucket : str
    prefix : str
        Prefix is the prefix of the shared object keys.

        Note: that within a bucket, the hierarchical key derivation scheme is
        delineated by forward slashes (/), so encryption information will be
        included in the resulting access grant to decrypt any key that shares
        the same prefix up until the last slash.

    Methods
    -------
    get_structure():
        _SharePrefixStruct
    """

    def __init__(self, bucket: str = "", prefix: str = ""):
        """Constructs all the necessary attributes for the SharePrefix object."""

        self.bucket = bucket
        self.prefix = prefix

    def get_structure(self):
        """Converts python class object to ctypes structure _SharePrefixStruct"""

        return _SharePrefixStruct(ctypes.c_char_p(self.bucket.encode('utf-8')),
                                  ctypes.c_char_p(self.prefix.encode('utf-8')))


class Bucket:
    """
    Bucket contains information about the bucket.

    ...

    Attributes
    ----------
    name : str
    created : int

    Methods
    -------
    get_structure():
        _BucketStruct
    get_dict():
        converts python class object to python dictionary
    """

    def __init__(self, name: str = "", created: int = 0):
        """Constructs all the necessary attributes for the Bucket object."""

        self.name = name
        self.created = created

    def get_structure(self):
        """Converts python class object to ctypes structure _BucketStruct"""

        return _BucketStruct(ctypes.c_char_p(self.name.encode('utf-8')),
                             ctypes.c_int64(self.created))

    def get_dict(self):
        """Converts python class object to python dictionary"""

        return {"name": self.name, "created": self.created}


class SystemMetadata:
    """
    SystemMetadata contains information about the object that cannot be changed directly.

    ...

    Attributes
    ----------
    created : int
    expires : int
    content_length : int

    Methods
    -------
    get_structure():
        _SystemMetadataStruct
    get_dict():
        converts python class object to python dictionary
    """

    def __init__(self, created: int = 0, expires: int = 0, content_length: int = 0):
        """Constructs all the necessary attributes for the SystemMetadata object."""

        self.created = created
        self.expires = expires
        self.content_length = content_length

    def get_structure(self):
        """Converts python class object to ctypes structure _SystemMetadataStruct"""

        return _SystemMetadataStruct(ctypes.c_int64(self.created),
                                     ctypes.c_int64(self.expires),
                                     ctypes.c_int64(self.content_length))

    def get_dict(self):
        """Converts python class object to python dictionary"""

        return {"created": self.created, "expires": self.expires,
                "content_length": self.content_length}


class CustomMetadataEntry:
    """
    CustomMetadata contains custom user metadata about the object.

    When choosing a custom key for your application start it with a prefix "app:key",
    as an example application named"Image Board" might use a key "image-board:title".

    ...

    Attributes
    ----------
    key : str
    key_length : int
    value : str
    value_length : int

    Methods
    -------
    get_structure():
        _CustomMetadataEntryStruct
    get_dict():
        converts python class object to python dictionary
    """

    def __init__(self, key: str = "", key_length: int = 0, value: str = "", value_length: int = 0):
        """Constructs all the necessary attributes for the CustomMetadataEntry object."""

        self.key = key
        self.key_length = key_length
        self.value = value
        self.value_length = value_length

    def get_structure(self):
        """Converts python class object to ctypes structure _CustomMetadataEntryStruct"""

        return _CustomMetadataEntryStruct(ctypes.c_char_p(self.key.encode('utf-8')),
                                          ctypes.c_size_t(self.key_length),
                                          ctypes.c_char_p(self.value.encode('utf-8')),
                                          ctypes.c_size_t(self.value_length))

    def get_dict(self):
        """Converts python class object to python dictionary"""

        return {"key": self.key, "key_length": self.key_length, "value": self.value,
                "value_length": self.value_length}


class CustomMetadata:
    """
    CustomMetadata contains a list of CustomMetadataEntry about the object.

    ...

    Attributes
    ----------
    entries : list of CustomMetadataEntry
    count : int

    Methods
    -------
    get_structure():
        _CustomMetadataStruct
    get_dict():
        converts python class object to python dictionary
    """

    def __init__(self, entries: [CustomMetadataEntry] = None, count: int = 0):
        """Constructs all the necessary attributes for the CustomMetadata object."""

        self.entries = entries
        self.count = count

    def get_structure(self):
        """Converts python class object to ctypes structure _CustomMetadataStruct"""

        if self.entries is None or self.count == 0:
            self.count = 0
            entries = ctypes.POINTER(_CustomMetadataEntryStruct)()
        else:
            li_array_size = (_CustomMetadataEntryStruct * self.count)()
            entries = ctypes.cast(li_array_size, ctypes.POINTER(_CustomMetadataEntryStruct))
            for i, val in enumerate(self.entries):
                entries[i] = val.get_structure()

        return _CustomMetadataStruct(entries, ctypes.c_size_t(self.count))

    def get_dict(self):
        """Converts python class object to python dictionary"""

        entries = self.entries
        if entries is None or self.count == 0:
            self.count = 0
            entries = [CustomMetadataEntry()]
        return {"entries": [entry.get_dict() for entry in entries], "count": self.count}


class Object:
    """
    Object contains information about an object.

    ...

    Attributes
    ----------
    key : str
    is_prefix : bool
        is_prefix indicates whether the Key is a prefix for other objects.
    system : SystemMetadata
    custom : CustomMetadata

    Methods
    -------
    get_structure():
        _ObjectStruct
    get_dict():
        converts python class object to python dictionary
    """

    def __init__(self, key: str = "", is_prefix: bool = False, system: SystemMetadata = None,
                 custom: CustomMetadata = None):
        """Constructs all the necessary attributes for the Object object."""

        self.key = key
        self.is_prefix = is_prefix
        self.system = system
        self.custom = custom

    def get_structure(self):
        """Converts python class object to ctypes structure _ObjectStruct"""

        if self.system is None:
            system = _SystemMetadataStruct()
        else:
            system = self.system.get_structure()

        if self.custom is None:
            custom = _CustomMetadataStruct()
        else:
            custom = self.custom.get_structure()

        return _ObjectStruct(ctypes.c_char_p(self.key.encode('utf-8')),
                             ctypes.c_bool(self.is_prefix), system, custom)

    def get_dict(self):
        """Converts python class object to python dictionary"""

        system = self.system
        custom = self.custom
        if system is None:
            system = SystemMetadata()
        if custom is None:
            custom = CustomMetadata()
        return {"key": self.key, "is_prefix": self.is_prefix, "system": system.get_dict(),
                "custom": custom.get_dict()}


class ListObjectsOptions:
    """
    ListObjectsOptions defines object listing options.

    ...

    Attributes
    ----------
    prefix : str
        prefix allows to filter objects by a key prefix. If not empty,
        it must end with slash.
    cursor : str
        cursor sets the starting position of the iterator.
        The first item listed will be the one after the cursor.
    recursive : bool
        recursive iterates the objects without collapsing prefixes.
    system : bool
        system includes SystemMetadata in the results.
    custom : bool
        custom includes CustomMetadata in the results.

    Methods
    -------
    get_structure():
        _ListObjectsOptionsStruct
    """

    def __init__(self, prefix: str = "", cursor: str = "", recursive: bool = False,
                 system: bool = False, custom: bool = False):
        """Constructs all the necessary attributes for the ListObjectsOptions object."""

        self.prefix = prefix
        self.cursor = cursor
        self.recursive = recursive
        self.system = system
        self.custom = custom

    def get_structure(self):
        """Converts python class object to ctypes structure _ListObjectsOptionsStruct"""

        return _ListObjectsOptionsStruct(ctypes.c_char_p(self.prefix.encode('utf-8')),
                                         ctypes.c_char_p(self.cursor.encode('utf-8')),
                                         ctypes.c_bool(self.recursive),
                                         ctypes.c_bool(self.system),
                                         ctypes.c_bool(self.custom))


class ListBucketsOptions:
    """
    ListBucketsOptions defines bucket listing options.

    ...

    Attributes
    ----------
    cursor : str
        Cursor sets the starting position of the iterator.
        The first item listed will be the one after the cursor.

    Methods
    -------
    get_structure():
        _ListBucketsOptionsStruct
    """

    def __init__(self, cursor: str = ""):
        """Constructs all the necessary attributes for the ListBucketsOptions object."""

        self.cursor = cursor

    def get_structure(self):
        """Converts python class object to ctypes structure _ListBucketsOptionsStruct"""

        return _ListBucketsOptionsStruct(ctypes.c_char_p(self.cursor.encode('utf-8')))


class UploadOptions:
    """
    UploadOptions contains additional options for uploading.

    ...

    Attributes
    ----------
    expires : int
        When expires is 0 or negative, it means no expiration.

    Methods
    -------
    get_structure():
        _UploadOptionsStruct
    """

    def __init__(self, expires: int):
        """Constructs all the necessary attributes for the UploadOptions object."""

        self.expires = expires

    def get_structure(self):
        """Converts python class object to ctypes structure _UploadOptionsStruct"""

        return _UploadOptionsStruct(ctypes.c_int64(self.expires))


class DownloadOptions:
    """
    DownloadOptions contains additional options for downloading.

    ...

    Attributes
    ----------
    offset : int
    length : int
        When length is negative, it will read until the end of the blob.

    Methods
    -------
    get_structure():
        _DownloadOptionsStruct
    """

    def __init__(self, offset: int, length: int):
        """Constructs all the necessary attributes for the DownloadOptions object."""

        self.offset = offset
        self.length = length

    def get_structure(self):
        """Converts python class object to ctypes structure _DownloadOptionsStruct"""

        return _DownloadOptionsStruct(ctypes.c_int64(self.offset),
                                      ctypes.c_int64(self.length))



"""Python user-defined exceptions for uplink errors"""

ERROR_INTERNAL = 0x02
ERROR_CANCELED = 0x03
ERROR_INVALID_HANDLE = 0x04
ERROR_TOO_MANY_REQUESTS = 0x05
ERROR_BANDWIDTH_LIMIT_EXCEEDED = 0x06

ERROR_BUCKET_NAME_INVALID = 0x10
ERROR_BUCKET_ALREADY_EXISTS = 0x11
ERROR_BUCKET_NOT_EMPTY = 0x12
ERROR_BUCKET_NOT_FOUND = 0x13

ERROR_OBJECT_KEY_INVALID = 0x20
ERROR_OBJECT_NOT_FOUND = 0x21
ERROR_UPLOAD_DONE = 0x22
ERROR_LIBUPLINK_SO_NOT_FOUND = 0x9999
"""_Error defines"""


class StorjException(Exception):
    """Base class for other exceptions

    Attributes:
        code -- error code
        message -- error message
        details -- error message from uplink-c
    """

    def __init__(self, message, code, details):
        self.message = message
        self.code = code
        self.details = details
        super(StorjException, self).__init__()

    def __str__(self):
        return repr(self.message)


class InternalError(StorjException):
    """Exception raised if internal error occurred.

    Attributes:
        details -- error message from uplink-c
    """

    def __init__(self, details):
        super().__init__("internal error", ERROR_INTERNAL, details)


class CancelledError(StorjException):
    """Exception raised if operation cancelled.

    Attributes:
        details -- error message from uplink-c
    """

    def __init__(self, details):
        super().__init__("operation canceled", ERROR_CANCELED, details)


class InvalidHandleError(StorjException):
    """Exception raised if handle is invalid.

    Attributes:
        details -- error message from uplink-c
    """

    def __init__(self, details):
        super().__init__("invalid handle", ERROR_INVALID_HANDLE, details)


class TooManyRequestsError(StorjException):
    """Exception raised if too many requests performed.

    Attributes:
        details -- error message from uplink-c
    """

    def __init__(self, details):
        super().__init__("too many requests", ERROR_TOO_MANY_REQUESTS, details)


class BandwidthLimitExceededError(StorjException):
    """Exception raised if allowed bandwidth limit exceeded.

    Attributes:
        details -- error message from uplink-c
    """

    def __init__(self, details):
        super().__init__("bandwidth limit exceeded", ERROR_BANDWIDTH_LIMIT_EXCEEDED, details)


class BucketNameInvalidError(StorjException):
    """Exception raised if bucket name is invalid.

    Attributes:
        details -- error message from uplink-c
    """

    def __init__(self, details):
        super().__init__("invalid bucket name", ERROR_BUCKET_NAME_INVALID, details)


class BucketAlreadyExistError(StorjException):
    """Exception raised if bucket already exist.

    Attributes:
        details -- error message from uplink-c
    """

    def __init__(self, details):
        super().__init__("bucket already exists", ERROR_BUCKET_ALREADY_EXISTS, details)


class BucketNotEmptyError(StorjException):
    """Exception raised if bucket is not empty.

    Attributes:
        details -- error message from uplink-c
    """

    def __init__(self, details):
        super().__init__("bucket is not empty", ERROR_BUCKET_NOT_EMPTY, details)


class BucketNotFoundError(StorjException):
    """Exception raised if bucket not found.

    Attributes:
        details -- error message from uplink-c
    """

    def __init__(self, details):
        super().__init__("bucket not found", ERROR_BUCKET_NOT_FOUND, details)


class ObjectKeyInvalidError(StorjException):
    """Exception raised if object key is invalid.

    Attributes:
        details -- error message from uplink-c
    """

    def __init__(self, details):
        super().__init__("invalid object key", ERROR_OBJECT_KEY_INVALID, details)


class ObjectNotFoundError(StorjException):
    """Exception raised if object not found.

    Attributes:
        details -- error message from uplink-c
    """

    def __init__(self, details):
        super().__init__("object not found", ERROR_OBJECT_NOT_FOUND, details)


class UploadDoneError(StorjException):
    """Exception raised if upload is complete.

    Attributes:
        details -- error message from uplink-c
    """

    def __init__(self, details):
        super().__init__("upload completed", ERROR_UPLOAD_DONE, details)


class LibUplinkSoError(StorjException):
    """Exception raised if reason is unknown.

    Attributes:
        code -- error code
        details -- error message from uplink-c
    """

    def __init__(self):
        super().__init__("libuplinkc.so not found", ERROR_LIBUPLINK_SO_NOT_FOUND,
                         "Please follow \"https://github.com/storj-thirdparty"
                         "/uplink-python#option-2\" "
                         "to build libuplinkc.so manually.")


def _storj_exception(code, details):
    switcher = {
        ERROR_INTERNAL: InternalError,
        ERROR_CANCELED: CancelledError,
        ERROR_INVALID_HANDLE: InvalidHandleError,
        ERROR_TOO_MANY_REQUESTS: TooManyRequestsError,
        ERROR_BANDWIDTH_LIMIT_EXCEEDED: BandwidthLimitExceededError,
        ERROR_BUCKET_NAME_INVALID: BucketNameInvalidError,
        ERROR_BUCKET_ALREADY_EXISTS: BucketAlreadyExistError,
        ERROR_BUCKET_NOT_EMPTY: BucketNotEmptyError,
        ERROR_BUCKET_NOT_FOUND: BucketNotFoundError,
        ERROR_OBJECT_KEY_INVALID: ObjectKeyInvalidError,
        ERROR_OBJECT_NOT_FOUND: ObjectNotFoundError,
        ERROR_UPLOAD_DONE: UploadDoneError
    }
    return switcher.get(code, StorjException)(details=details)



# pylint: disable=missing-docstring, broad-except
import subprocess
import os
import platform
import sysconfig

import setuptools
from setuptools.command.install import install

with open("README.md", "r") as fh:
    long_description = fh.read()


class Install(install):

    @staticmethod
    def find_module_path():
        new_path = os.path.join(sysconfig.get_paths()['purelib'], "uplink_python")
        try:
            os.makedirs(new_path, exist_ok=True)
            os.system("echo Directory uplink_python created successfully.")
        except OSError as error:
            os.system("echo Error in creating uplink_python directory. Error: "+str(error))
        return new_path

    def run(self):

        try:
            install_path = self.find_module_path()
            os.system("echo Package installation path: " + install_path)
            if platform.system() == "Windows":
                os.system("icacls " + install_path + " /grant Everyone:F /t")
            else:
                os.system("sudo chmod -R 777 " + install_path)
            os.system("echo Building libuplinkc.so")
            copy_command = "copy" if platform.system() == "Windows" else "cp"
            command = "git clone https://github.com/storj/uplink-c && cd uplink-c" \
                      "&& go build -o libuplinkc.so " \
                      "-buildmode=c-shared && " + copy_command + " *.so " + install_path
            build_so = subprocess.Popen(command,
                                        stdout=subprocess.PIPE,
                                        stderr=subprocess.STDOUT, shell=True)
            output, errors = build_so.communicate()
            build_so.wait()
            if output is not None:
                os.system("echo " + output.decode('utf-8'))
                os.system("echo Building libuplinkc.so successful.")
            if errors is not None:
                os.system("echo " + errors.decode('utf-8'))
                os.system("echo Building libuplinkc.so failed.")

        except Exception as error:
            os.system("echo " + str(error))
            os.system("echo Building libuplinkc.so failed.")

        install.run(self)


setuptools.setup(
    name="uplink-python",
    version="1.0.5",
    author="Utropicmedia",
    author_email="development@utropicmedia.com",
    license='Apache Software License',
    description="Python-native language binding for uplink to"
                "communicate with the Storj network.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/storj-thirdparty/uplink-python",

    packages=['uplink_python'],
    install_requires=['wheel'],
    include_package_data=True,
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
        "Topic :: Software Development :: Build Tools",
    ],
    python_requires='>=3.4',
    cmdclass={
        'install': Install,
    }
)



"""C-type Classes for input and output interaction with libuplinkc."""
# pylint: disable=too-few-public-methods
import ctypes


class _ConfigStruct(ctypes.Structure):
    """Config ctypes structure for internal processing."""

    _fields_ = [("user_agent", ctypes.c_char_p), ("dial_timeout_milliseconds", ctypes.c_int32),
                ("temp_directory", ctypes.c_char_p)]


class _AccessStruct(ctypes.Structure):
    """Access ctypes structure for internal processing."""

    _fields_ = [("_handle", ctypes.c_size_t)]


class _PermissionStruct(ctypes.Structure):
    """Permission ctypes structure for internal processing."""

    _fields_ = [("allow_download", ctypes.c_bool), ("allow_upload", ctypes.c_bool),
                ("allow_list", ctypes.c_bool), ("allow_delete", ctypes.c_bool),
                ("not_before", ctypes.c_int64), ("not_after", ctypes.c_int64)]


class _SharePrefixStruct(ctypes.Structure):
    """SharePrefix ctypes structure for internal processing."""

    _fields_ = [("bucket", ctypes.c_char_p), ("prefix", ctypes.c_char_p)]


class _BucketStruct(ctypes.Structure):
    """Bucket ctypes structure for internal processing."""

    _fields_ = [("name", ctypes.c_char_p), ("created", ctypes.c_int64)]


class _ProjectStruct(ctypes.Structure):
    """Project ctypes structure for internal processing."""

    _fields_ = [("_handle", ctypes.c_size_t)]


class _SystemMetadataStruct(ctypes.Structure):
    """SystemMetadata ctypes structure for internal processing."""

    _fields_ = [("created", ctypes.c_int64), ("expires", ctypes.c_int64),
                ("content_length", ctypes.c_int64)]


class _CustomMetadataEntryStruct(ctypes.Structure):
    """CustomMetadataEntry ctypes structure for internal processing."""

    _fields_ = [("key", ctypes.c_char_p), ("key_length", ctypes.c_size_t),
                ("value", ctypes.c_char_p), ("value_length", ctypes.c_size_t)]


class _CustomMetadataStruct(ctypes.Structure):
    """CustomMetadata ctypes structure for internal processing."""

    _fields_ = [("entries", ctypes.POINTER(_CustomMetadataEntryStruct)), ("count", ctypes.c_size_t)]


class _ObjectStruct(ctypes.Structure):
    """Object ctypes structure for internal processing."""

    _fields_ = [("key", ctypes.c_char_p), ("is_prefix", ctypes.c_bool),
                ("system", _SystemMetadataStruct), ("custom", _CustomMetadataStruct)]


class _ListObjectsOptionsStruct(ctypes.Structure):
    """ListObjectsOptions ctypes structure for internal processing."""

    _fields_ = [("prefix", ctypes.c_char_p), ("cursor", ctypes.c_char_p),
                ("recursive", ctypes.c_bool), ("system", ctypes.c_bool), ("custom", ctypes.c_bool)]


class _ListBucketsOptionsStruct(ctypes.Structure):
    """ListBucketsOptions ctypes structure for internal processing."""

    _fields_ = [("cursor", ctypes.c_char_p)]


class _ObjectIterator(ctypes.Structure):
    """ObjectIterator ctypes structure"""

    _fields_ = [("_handle", ctypes.c_size_t)]


class _BucketIterator(ctypes.Structure):
    """BucketIterator ctypes structure for internal processing"""

    _fields_ = [("_handle", ctypes.c_size_t)]


class _UploadStruct(ctypes.Structure):
    """Upload ctypes structure for internal processing."""

    _fields_ = [("_handle", ctypes.c_size_t)]


class _UploadOptionsStruct(ctypes.Structure):
    """UploadOptions ctypes structure for internal processing."""

    _fields_ = [("expires", ctypes.c_int64)]


class _DownloadStruct(ctypes.Structure):
    """Download ctypes structure for internal processing."""

    _fields_ = [("_handle", ctypes.c_size_t)]


class _DownloadOptionsStruct(ctypes.Structure):
    """DownloadOptions ctypes structure for internal processing."""

    _fields_ = [("offset", ctypes.c_int64), ("length", ctypes.c_int64)]


class _Error(ctypes.Structure):
    """Error ctypes structure for internal processing."""

    _fields_ = [("code", ctypes.c_int32), ("message", ctypes.c_char_p)]


class _ProjectResult(ctypes.Structure):
    """ProjectResult ctypes structure"""

    _fields_ = [("project", ctypes.POINTER(_ProjectStruct)), ("error", ctypes.POINTER(_Error))]


class _BucketResult(ctypes.Structure):
    """BucketResult ctypes structure"""

    _fields_ = [("bucket", ctypes.POINTER(_BucketStruct)), ("error", ctypes.POINTER(_Error))]


class _UploadResult(ctypes.Structure):
    """UploadResult ctypes structure"""

    _fields_ = [("upload", ctypes.POINTER(_UploadStruct)), ("error", ctypes.POINTER(_Error))]


class _DownloadResult(ctypes.Structure):
    """DownloadResult ctypes structure"""

    _fields_ = [("download", ctypes.POINTER(_DownloadStruct)), ("error", ctypes.POINTER(_Error))]


class _AccessResult(ctypes.Structure):
    """AccessResult ctypes structure"""

    _fields_ = [("access", ctypes.POINTER(_AccessStruct)), ("error", ctypes.POINTER(_Error))]


class _StringResult(ctypes.Structure):
    """StringResult ctypes structure"""

    _fields_ = [("string", ctypes.c_char_p), ("error", ctypes.POINTER(_Error))]


class _ObjectResult(ctypes.Structure):
    """ObjectResult ctypes structure"""

    _fields_ = [("object", ctypes.POINTER(_ObjectStruct)), ("error", ctypes.POINTER(_Error))]


class _WriteResult(ctypes.Structure):
    """WriteResult ctypes structure"""

    _fields_ = [("bytes_written", ctypes.c_size_t), ("error", ctypes.POINTER(_Error))]


class _ReadResult(ctypes.Structure):
    """ReadResult ctypes structure"""

    _fields_ = [("bytes_read", ctypes.c_size_t), ("error", ctypes.POINTER(_Error))]



"""Module with Project class and project methods to work with buckets and objects"""
import ctypes

from uplink_python.module_classes import ListBucketsOptions, ListObjectsOptions, UploadOptions, DownloadOptions
from uplink_python.module_def import _BucketStruct, _ObjectStruct, _ListObjectsOptionsStruct, _ObjectResult, \
    _ListBucketsOptionsStruct, _UploadOptionsStruct, _DownloadOptionsStruct, _ProjectStruct, \
    _BucketResult, _BucketIterator, _ObjectIterator, _DownloadResult, _UploadResult, _Error
from uplink_python.upload import Upload
from uplink_python.download import Download
from uplink_python.errors import _storj_exception


class Project:
    """
    Project provides access to managing buckets and objects.

    ...

    Attributes
    ----------
    project : int
        Project _handle returned from libuplinkc project_result.project
    uplink : Uplink
        uplink object used to get access

    Methods
    -------
    create_bucket():
        Bucket
    ensure_bucket():
        Bucket
    stat_bucket():
        Bucket
    list_buckets():
        list of Bucket
    delete_bucket():
        Bucket
    stat_object():
        Object
    list_objects():
        list of Object
    delete_object():
        Object
    close():
        None
    upload_object():
        Upload
    download_object():
        Download
    """

    def __init__(self, project, uplink):
        """Constructs all the necessary attributes for the Project object."""

        self.project = project
        self.uplink = uplink

    def create_bucket(self, bucket_name: str):
        """
        function creates a new bucket.
        When bucket already exists it throws BucketAlreadyExistError exception.

        Parameters
        ----------
        bucket_name : str

        Returns
        -------
        Bucket
        """

        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.create_bucket.argtypes = [ctypes.POINTER(_ProjectStruct),
                                                          ctypes.c_char_p]
        self.uplink.m_libuplink.create_bucket.restype = _BucketResult
        #
        # prepare the input for the function
        bucket_name_ptr = ctypes.c_char_p(bucket_name.encode('utf-8'))

        # create bucket by calling the exported golang function
        bucket_result = self.uplink.m_libuplink.create_bucket(self.project, bucket_name_ptr)
        #
        # if error occurred
        if bool(bucket_result.error):
            raise _storj_exception(bucket_result.error.contents.code,
                                   bucket_result.error.contents.message.decode("utf-8"))
        return self.uplink.bucket_from_result(bucket_result.bucket)

    def ensure_bucket(self, bucket_name: str):
        """
        function ensures that a bucket exists or creates a new one.

        When bucket already exists it returns a valid Bucket and no error

        Parameters
        ----------
        bucket_name : str

        Returns
        -------
        Bucket
        """

        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.ensure_bucket.argtypes = [ctypes.POINTER(_ProjectStruct),
                                                          ctypes.c_char_p]
        self.uplink.m_libuplink.ensure_bucket.restype = _BucketResult
        #
        # prepare the input for the function
        bucket_name_ptr = ctypes.c_char_p(bucket_name.encode('utf-8'))

        # open bucket if doesn't exist by calling the exported golang function
        bucket_result = self.uplink.m_libuplink.ensure_bucket(self.project, bucket_name_ptr)
        #
        # if error occurred
        if bool(bucket_result.error):
            raise _storj_exception(bucket_result.error.contents.code,
                                   bucket_result.error.contents.message.decode("utf-8"))
        return self.uplink.bucket_from_result(bucket_result.bucket)

    def stat_bucket(self, bucket_name: str):
        """
        function returns information about a bucket.

        Parameters
        ----------
        bucket_name : str

        Returns
        -------
        Bucket
        """

        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.stat_bucket.argtypes = [ctypes.POINTER(_ProjectStruct),
                                                        ctypes.c_char_p]
        self.uplink.m_libuplink.stat_bucket.restype = _BucketResult
        #
        # prepare the input for the function
        bucket_name_ptr = ctypes.c_char_p(bucket_name.encode('utf-8'))

        # get bucket information by calling the exported golang function
        bucket_result = self.uplink.m_libuplink.stat_bucket(self.project, bucket_name_ptr)
        #
        # if error occurred
        if bool(bucket_result.error):
            raise _storj_exception(bucket_result.error.contents.code,
                                   bucket_result.error.contents.message.decode("utf-8"))
        return self.uplink.bucket_from_result(bucket_result.bucket)

    def list_buckets(self, list_bucket_options: ListBucketsOptions = None):
        """
        function returns a list of buckets with all its information.

        Parameters
        ----------
        list_bucket_options : ListBucketsOptions (optional)

        Returns
        -------
        list of Bucket
        """

        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.list_buckets.argtypes = [ctypes.POINTER(_ProjectStruct),
                                                         ctypes.POINTER(_ListBucketsOptionsStruct)]
        self.uplink.m_libuplink.list_buckets.restype = ctypes.POINTER(_BucketIterator)
        #
        self.uplink.m_libuplink.bucket_iterator_item.argtypes = [ctypes.POINTER(_BucketIterator)]
        self.uplink.m_libuplink.bucket_iterator_item.restype = ctypes.POINTER(_BucketStruct)
        #
        self.uplink.m_libuplink.bucket_iterator_err.argtypes = [ctypes.POINTER(_BucketIterator)]
        self.uplink.m_libuplink.bucket_iterator_err.restype = ctypes.POINTER(_Error)
        #
        self.uplink.m_libuplink.bucket_iterator_next.argtypes = [ctypes.POINTER(_BucketIterator)]
        self.uplink.m_libuplink.bucket_iterator_next.restype = ctypes.c_bool
        #
        # prepare the input for the function
        if list_bucket_options is None:
            list_bucket_options_obj = ctypes.POINTER(_ListBucketsOptionsStruct)()
        else:
            list_bucket_options_obj = ctypes.byref(list_bucket_options.get_structure())

        # get bucket list by calling the exported golang function
        bucket_iterator = self.uplink.m_libuplink.list_buckets(self.project,
                                                               list_bucket_options_obj)

        bucket_iterator_err = self.uplink.m_libuplink.bucket_iterator_err(bucket_iterator)
        if bool(bucket_iterator_err):
            raise _storj_exception(bucket_iterator_err.contents.code,
                                   bucket_iterator_err.contents.message.decode("utf-8"))

        bucket_list = list()
        while self.uplink.m_libuplink.bucket_iterator_next(bucket_iterator):
            bucket = self.uplink.m_libuplink.bucket_iterator_item(bucket_iterator)
            bucket_list.append(self.uplink.bucket_from_result(bucket))

        return bucket_list

    def delete_bucket(self, bucket_name: str):
        """
        function deletes a bucket.

        When bucket is not empty it throws BucketNotEmptyError exception.

        Parameters
        ----------
        bucket_name : str

        Returns
        -------
        Bucket
        """

        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.delete_bucket.argtypes = [ctypes.POINTER(_ProjectStruct),
                                                          ctypes.c_char_p]
        self.uplink.m_libuplink.delete_bucket.restype = _BucketResult
        #
        # prepare the input for the function
        bucket_name_ptr = ctypes.c_char_p(bucket_name.encode('utf-8'))

        # delete bucket by calling the exported golang function
        bucket_result = self.uplink.m_libuplink.delete_bucket(self.project, bucket_name_ptr)
        #
        # if error occurred
        if bool(bucket_result.error):
            raise _storj_exception(bucket_result.error.contents.code,
                                   bucket_result.error.contents.message.decode("utf-8"))
        return self.uplink.bucket_from_result(bucket_result.bucket)

    def stat_object(self, bucket_name: str, storj_path: str):
        """
        function returns information about an object at the specific key.

        Parameters
        ----------
        bucket_name : str
        storj_path : str

        Returns
        -------
        Object
        """

        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.stat_object.argtypes = [ctypes.POINTER(_ProjectStruct),
                                                        ctypes.c_char_p, ctypes.c_char_p]
        self.uplink.m_libuplink.stat_object.restype = _ObjectResult
        #
        # prepare the input for the function
        bucket_name_ptr = ctypes.c_char_p(bucket_name.encode('utf-8'))
        storj_path_ptr = ctypes.c_char_p(storj_path.encode('utf-8'))

        # get object information by calling the exported golang function
        object_result = self.uplink.m_libuplink.stat_object(self.project, bucket_name_ptr,
                                                            storj_path_ptr)
        #
        # if error occurred
        if bool(object_result.error):
            raise _storj_exception(object_result.error.contents.code,
                                   object_result.error.contents.message.decode("utf-8"))
        return self.uplink.object_from_result(object_result.object)

    def list_objects(self, bucket_name: str, list_object_options: ListObjectsOptions = None):
        """
        function returns a list of objects with all its information.

        Parameters
        ----------
        bucket_name : str
        list_object_options : ListObjectsOptions (optional)

        Returns
        -------
        list of Object
        """

        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.list_objects.argtypes = [ctypes.POINTER(_ProjectStruct),
                                                         ctypes.c_char_p,
                                                         ctypes.POINTER(_ListObjectsOptionsStruct)]
        self.uplink.m_libuplink.list_objects.restype = ctypes.POINTER(_ObjectIterator)
        #
        self.uplink.m_libuplink.object_iterator_item.argtypes = [ctypes.POINTER(_ObjectIterator)]
        self.uplink.m_libuplink.object_iterator_item.restype = ctypes.POINTER(_ObjectStruct)
        #
        self.uplink.m_libuplink.object_iterator_err.argtypes = [ctypes.POINTER(_ObjectIterator)]
        self.uplink.m_libuplink.object_iterator_err.restype = ctypes.POINTER(_Error)
        #
        self.uplink.m_libuplink.object_iterator_next.argtypes = [ctypes.POINTER(_ObjectIterator)]
        self.uplink.m_libuplink.object_iterator_next.restype = ctypes.c_bool
        #
        # prepare the input for the function
        if list_object_options is None:
            list_object_options_obj = ctypes.POINTER(_ListObjectsOptionsStruct)()
        else:
            list_object_options_obj = ctypes.byref(list_object_options.get_structure())
        bucket_name_ptr = ctypes.c_char_p(bucket_name.encode('utf-8'))

        # get object list by calling the exported golang function
        object_iterator = self.uplink.m_libuplink.list_objects(self.project, bucket_name_ptr,
                                                               list_object_options_obj)

        object_iterator_err = self.uplink.m_libuplink.object_iterator_err(object_iterator)
        if bool(object_iterator_err):
            raise _storj_exception(object_iterator_err.contents.code,
                                   object_iterator_err.contents.message.decode("utf-8"))

        object_list = list()
        while self.uplink.m_libuplink.object_iterator_next(object_iterator):
            object_ = self.uplink.m_libuplink.object_iterator_item(object_iterator)
            object_list.append(self.uplink.object_from_result(object_))
        return object_list

    def delete_object(self, bucket_name: str, storj_path: str):
        """
        function deletes the object at the specific key.

        Parameters
        ----------
        bucket_name : str
        storj_path : str

        Returns
        -------
        Object
        """

        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.delete_object.argtypes = [ctypes.POINTER(_ProjectStruct),
                                                          ctypes.c_char_p, ctypes.c_char_p]
        self.uplink.m_libuplink.delete_object.restype = _ObjectResult
        #
        # prepare the input for the function
        bucket_name_ptr = ctypes.c_char_p(bucket_name.encode('utf-8'))
        storj_path_ptr = ctypes.c_char_p(storj_path.encode('utf-8'))

        # delete object by calling the exported golang function
        object_result = self.uplink.m_libuplink.delete_object(self.project, bucket_name_ptr,
                                                              storj_path_ptr)
        #
        # if error occurred
        if bool(object_result.error):
            raise _storj_exception(object_result.error.contents.code,
                                   object_result.error.contents.message.decode("utf-8"))
        return self.uplink.object_from_result(object_result.object)

    def close(self):
        """
        function closes the project and all associated resources.

        Returns
        -------
        None
        """
        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.close_project.argtypes = [ctypes.POINTER(_ProjectStruct)]
        self.uplink.m_libuplink.close_project.restype = ctypes.POINTER(_Error)
        #
        # close Storj project by calling the exported golang function
        error = self.uplink.m_libuplink.close_project(self.project)
        #
        # if error occurred
        if bool(error):
            raise _storj_exception(error.contents.code,
                                   error.contents.message.decode("utf-8"))

    def upload_object(self, bucket_name: str, storj_path: str,
                      upload_options: UploadOptions = None):
        """
        function starts an upload to the specified key.

        Parameters
        ----------
        bucket_name : str
        storj_path : str
        upload_options : UploadOptions (optional)

        Returns
        -------
        Upload
        """
        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.upload_object.argtypes = [ctypes.POINTER(_ProjectStruct),
                                                          ctypes.c_char_p, ctypes.c_char_p,
                                                          ctypes.POINTER(_UploadOptionsStruct)]
        self.uplink.m_libuplink.upload_object.restype = _UploadResult
        #
        # prepare the input for the function
        if upload_options is None:
            upload_options_obj = ctypes.POINTER(_UploadOptionsStruct)()
        else:
            upload_options_obj = ctypes.byref(upload_options.get_structure())

        bucket_name_ptr = ctypes.c_char_p(bucket_name.encode('utf-8'))
        storj_path_ptr = ctypes.c_char_p(storj_path.encode('utf-8'))

        # get uploader by calling the exported golang function
        upload_result = self.uplink.m_libuplink.upload_object(self.project, bucket_name_ptr,
                                                              storj_path_ptr,
                                                              upload_options_obj)
        #
        # if error occurred
        if bool(upload_result.error):
            raise _storj_exception(upload_result.error.contents.code,
                                   upload_result.error.contents.message.decode("utf-8"))
        return Upload(upload_result.upload, self.uplink)

    def download_object(self, bucket_name: str, storj_path: str,
                        download_options: DownloadOptions = None):
        """
        function starts download to the specified key.

        Parameters
        ----------
        bucket_name : str
        storj_path : str
        download_options : DownloadOptions (optional)

        Returns
        -------
        Download
        """
        #
        # declare types of arguments and response of the corresponding golang function
        self.uplink.m_libuplink.download_object.argtypes = [ctypes.POINTER(_ProjectStruct),
                                                            ctypes.c_char_p, ctypes.c_char_p,
                                                            ctypes.POINTER(_DownloadOptionsStruct)]
        self.uplink.m_libuplink.download_object.restype = _DownloadResult
        #
        # prepare the input for the function
        if download_options is None:
            download_options_obj = ctypes.POINTER(_DownloadOptionsStruct)()
        else:
            download_options_obj = ctypes.byref(download_options.get_structure())

        bucket_name_ptr = ctypes.c_char_p(bucket_name.encode('utf-8'))
        storj_path_ptr = ctypes.c_char_p(storj_path.encode('utf-8'))

        # get downloader by calling the exported golang function
        download_result = self.uplink.m_libuplink.download_object(self.project, bucket_name_ptr,
                                                                  storj_path_ptr,
                                                                  download_options_obj)
        #
        # if error occurred
        if bool(download_result.error):
            raise _storj_exception(download_result.error.contents.code,
                                   download_result.error.contents.message.decode("utf-8"))
        return Download(download_result.download, self.uplink, self.project, bucket_name_ptr,
                        storj_path_ptr)
